from datetime import time

import pandas as pd

from visions.backends.pandas.series_utils import (
    class_name_attrs,
    series_handle_nulls,
    series_not_empty,
)
from visions.types.time import Time

# @Time.register_relationship(DateTime, pd.Series)
# @series_handle_nulls
# def datetime_is_time(series: pd.Series) -> bool:
#     dtseries = series.dt.date
#     value = date(1, 1, 1)
#     return all(v == value for v in dtseries)
#
#
# @Time.register_transformer(DateTime, pd.Series)
# def datetime_to_time(series: pd.Series, state: dict) -> pd.Series:
#     return series.dt.time


@Time.contains_op.register
@series_handle_nulls
@series_not_empty
def time_contains(series: pd.Series, state: dict) -> bool:
    return class_name_attrs(series, time, ["microsecond", "hour"])



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.complex import Complex
from visions.types.generic import Generic
from visions.types.string import String
from visions.types.type import VisionsBaseType


class Float(VisionsBaseType):
    """**Float** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> import visions
        >>> x = [1.0, 2.5, 5.0]
        >>> x in visions.Float
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [
            IdentityRelation(Generic),
            InferenceRelation(String),
            InferenceRelation(Complex),
        ]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



from typing import Sequence

from visions.backends.python.series_utils import (
    sequence_handle_none,
    sequence_not_empty,
)
from visions.types.string import String


@String.contains_op.register
@sequence_not_empty
@sequence_handle_none
def string_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(v, str) for v in sequence)



from visions.types import (
    Boolean,
    Categorical,
    Complex,
    DateTime,
    Float,
    Generic,
    Integer,
    Object,
    String,
    TimeDelta,
)
from visions.typesets.typeset import VisionsTypeset


class StandardSet(VisionsTypeset):
    """The standard visions typesets

    Includes support for the following types:

    - Float
    - Integer
    - Boolean
    - Object
    - String
    - Complex
    - Categorical
    - DateTime
    - TimeDelta

    """

    def __init__(self) -> None:
        types = {
            Generic,
            Boolean,
            Float,
            Object,
            Complex,
            Categorical,
            DateTime,
            TimeDelta,
            Integer,
            String,
        }
        super().__init__(types)



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, TypeRelation
from visions.types.generic import Generic
from visions.types.type import VisionsBaseType


class TimeDelta(VisionsBaseType):
    """**TimeDelta** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> from datetime import timedelta
        >>> x = [timedelta(hours=1), timedelta(hours=3)]
        >>> x in visions.Timedelta
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [IdentityRelation(Generic)]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



import functools
import os
import sys
import warnings
from typing import Callable, TypeVar

T = TypeVar("T")


def suppress_warnings(func: Callable[..., T]) -> Callable[..., T]:
    """Suppress warnings produces while executing the wrapped function."""

    @functools.wraps(func)
    def inner(*args, **kwargs) -> T:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return func(*args, **kwargs)

    return inner


def discard_stderr(func: Callable[..., T]) -> Callable[..., T]:
    """Shapely logs failures at a silly severity, just trying to suppress it's output on failures.
    Only known way to get rid of sys output when wkt.loads hits a bad value"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> T:
        sys.stderr = open(os.devnull, "w")
        res = func(*args, **kwargs)
        sys.stderr = sys.__stderr__
        return res

    return wrapper



import numpy as np

from visions.backends.numpy.array_utils import array_not_empty
from visions.types.time_delta import TimeDelta


@TimeDelta.contains_op.register
@array_not_empty
def time_delta_contains(array: np.ndarray, state: dict) -> bool:
    """
    Example:
        >>> x = pd.array([pd.Timedelta(days=i) for i in range(3)])
        >>> x in visions.Timedelta
        True
    """
    return np.issubdtype(array.dtype, np.timedelta64)



from typing import Any, Callable, List, Optional, Sequence, Type, TypeVar, Union

from visions.relations import IdentityRelation, InferenceRelation
from visions.types.type import VisionsBaseType

T = TypeVar("T")


def process_relation(items: Union[dict, Type[VisionsBaseType]]) -> IdentityRelation:
    if isinstance(items, dict):
        return IdentityRelation(**items)
    elif issubclass(items, VisionsBaseType):
        return IdentityRelation(related_type=items)
    else:
        raise TypeError("identity should be a list, a dict of params or related_type.")


def create_type(
    name: str,
    contains: Callable[[Any, dict], bool],
    identity: Optional[
        Union[Type[VisionsBaseType], List[Union[dict, Type[VisionsBaseType]]], dict]
    ] = None,
    inference: Optional[Union[List[dict], dict]] = None,
):
    def get_relations():
        if isinstance(identity, Sequence):
            relations = [process_relation(item) for item in identity]
        else:
            relations = [] if identity is None else [process_relation(identity)]

        if inference is not None:
            if isinstance(inference, dict):
                relations += [InferenceRelation(**inference)]
            elif isinstance(inference, list):
                relations += [InferenceRelation(**params) for params in inference]
            else:
                raise TypeError("inference should be a list or a dict of params.")

        return relations

    def contains_op(series, state):
        return contains(series, state)

    return type(
        name,
        (VisionsBaseType,),
        {
            "get_relations": staticmethod(get_relations),
            "contains_op": staticmethod(contains_op),
        },
    )



import imghdr
import pathlib
from typing import Sequence

from visions.types.image import Image


@Image.contains_op.register
def image_contains(sequence: Sequence, state: dict) -> bool:
    return all(
        isinstance(p, pathlib.Path) and p.exists() and imghdr.what(p) for p in sequence
    )



import uuid
from typing import Sequence

from visions.types.string import String
from visions.types.uuid import UUID


@UUID.contains_op.register
def uuid_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(value, uuid.UUID) for value in sequence)


@UUID.register_transformer(String, Sequence)
def string_to_uuid(sequence: Sequence, state: dict) -> Sequence:
    return [uuid.UUID(value) for value in sequence]


@UUID.register_relationship(String, Sequence)
def string_is_uuid(sequence: Sequence, state: dict) -> bool:
    try:
        string_to_uuid(sequence)
        return True
    except (ValueError, TypeError, AttributeError):
        return False



from typing import Dict

import numpy as np
import pandas as pd

from visions.backends.pandas.test_utils import pandas_version

not_pandas_1_0_5 = not (
    (pandas_version[0] == 1) and (pandas_version[1] == 0) and (pandas_version[2] == 5)
)


def get_sparse_series() -> Dict[str, pd.Series]:
    test_series = {
        "int_sparse": pd.Series([-1, 0, 1, 2, 3], dtype=pd.SparseDtype(np.int32, 0)),
        "float_sparse": pd.Series(
            [np.nan, 0.2, 1, 2, 3],
            dtype=pd.SparseDtype(np.float64, np.nan),
        ),
        "complex_sparse": pd.Series(
            [np.nan, complex(0, 1), complex(1, -1), complex(2, 4), complex(3, -12)],
            dtype=pd.SparseDtype(np.complex128, np.nan),
        ),
        "bool_sparse": pd.Series(
            [True, False, False],
            dtype=pd.SparseDtype(np.bool_, False),
        ),
        "str_obj_sparse": pd.Series(
            pd.arrays.SparseArray([None, None, "gold", "black", "silver"]),
        ),
        # Pending https://github.com/pandas-dev/pandas/issues/35762
        # pd.Series([None, 0, 1, 2, 3, 4], name="datetime_sparse", dtype=pd.SparseDtype(np.datetime64)),
        # Pandas dtypes
        "pd_int64_sparse": pd.Series(
            [0, 1, 2, 3, None],
            dtype=pd.SparseDtype("int", np.nan),
        ),
        # Pending https://github.com/pandas-dev/pandas/issues/35793
        # pd.Series(
        #     ["a", "b", "c", None],
        #     name="pd_categorical_sparse",
        #     dtype=pd.SparseDtype(pd.CategoricalDtype(['a', 'b', 'c', 'd']))
        # )
    }

    if pandas_version[0] >= 1 and not_pandas_1_0_5:
        test_series["pd_string_sparse"] = pd.Series(
            ["Patty", "Valentine", "Upper", "", "", ""],
            dtype=pd.SparseDtype(pd.StringDtype(), ""),
        )
        test_series["pd_bool_sparse"] = pd.Series(
            [True, False, False, None],
            dtype=pd.SparseDtype("bool", pd.NA),
        )

    return test_series



import os
import sys
from typing import Sequence

from visions.types.geometry import Geometry
from visions.types.string import String


@Geometry.register_relationship(String, Sequence)
def string_is_geometry(sequence: Sequence, state: dict) -> bool:
    """Shapely logs failures at a silly severity, just trying to suppress it's output on failures."""
    from shapely import wkt
    from shapely.errors import WKTReadingError

    # only way to get rid of sys output when wkt.loads hits a bad value
    # TODO: use coercion wrapper for this
    sys.stderr = open(os.devnull, "w")
    try:
        result = all(wkt.loads(value) for value in sequence)
    except (WKTReadingError, AttributeError, UnicodeEncodeError, TypeError):
        result = False
    finally:
        sys.stderr = sys.__stderr__
    return result


@Geometry.register_transformer(String, Sequence)
def string_to_geometry(sequence: Sequence, state: dict) -> Sequence:
    from shapely import wkt

    return tuple(map(wkt.loads, sequence))


@Geometry.contains_op.register
def geometry_contains(sequence: Sequence, state: dict) -> bool:
    from shapely.geometry.base import BaseGeometry

    return all(issubclass(type(x), BaseGeometry) for x in sequence)



from typing import Dict

import pandas as pd


def get_geometry_series() -> Dict[str, pd.Series]:
    from shapely import wkt

    series = {
        "geometry_string_series": pd.Series(
            ["POINT (-92 42)", "POINT (-92 42.1)", "POINT (-92 42.2)"],
        ),
        "geometry_series": pd.Series(
            [
                wkt.loads("POINT (-92 42)"),
                wkt.loads("POINT (-92 42.1)"),
                wkt.loads("POINT (-92 42.2)"),
            ],
        ),
        "geometry_series_missing": pd.Series(
            [
                wkt.loads("POINT (-92 42)"),
                wkt.loads("POINT (-92 42.1)"),
                wkt.loads("POINT (-92 42.2)"),
                None,
            ],
        ),
    }
    return series



import imghdr
from pathlib import Path

import pandas as pd

from visions.backends.pandas.series_utils import series_handle_nulls, series_not_empty
from visions.types.image import Image


@Image.contains_op.register
@series_not_empty
@series_handle_nulls
def image_contains(series: pd.Series, state: dict) -> bool:
    return all(isinstance(p, Path) and p.exists() and imghdr.what(p) for p in series)



# import pandas as pd
#
# from visions.backends.pandas_be import test_utils
# from visions.relations import InferenceRelation
# from visions.relations.string_to_datetime import to_datetime_year_month_day
# from visions.types import Integer
#
#
# def to_datetime(series: pd.Series) -> pd.Series:
#     return pd.to_datetime(series)
#
#
# def _to_datetime(func) -> InferenceRelation:
#     return InferenceRelation(
#         relationship=test_utils.coercion_test(lambda s: func(s.astype(str))),
#         transformer=to_datetime,
#         related_type=Integer,
#     )
#
#
# # TODO: do only convert obvious dates (20191003000000)
# def integer_to_datetime(cls):
#     return _to_datetime(cls, to_datetime)
#
#
# def integer_to_datetime_year_month_day(cls) -> InferenceRelation:
#     return _to_datetime(cls, to_datetime_year_month_day)



from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import ByteType, IntegerType, LongType, ShortType

from visions.types.integer import Integer


@Integer.contains_op.register
def integer_contains(sequence: DataFrame, state: dict) -> bool:
    if len(sequence.schema) != 1:
        return False

    dtype = sequence.schema[0].dataType
    return isinstance(dtype, (ByteType, ShortType, IntegerType, LongType))



import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas.series_utils import (
    series_handle_nulls,
    series_not_empty,
    series_not_sparse,
)
from visions.types.string import String

pandas_has_string_dtype_flag = hasattr(pdt, "is_string_dtype")


@series_handle_nulls
def _is_string(series: pd.Series, state: dict):
    if not all(isinstance(v, str) for v in series.values[0:5]):
        return False
    try:
        return (series.astype(str).values == series.values).all()
    except (TypeError, ValueError):
        return False


@String.contains_op.register
@series_not_sparse
@series_not_empty
def string_contains(series: pd.Series, state: dict) -> bool:
    if pdt.is_categorical_dtype(series):
        return False
    elif not pdt.is_object_dtype(series):
        return pandas_has_string_dtype_flag and pdt.is_string_dtype(series)

    return _is_string(series, state)



from typing import Dict, List

import numpy as np

from visions.backends.numpy.array_utils import (
    all_type,
    array_handle_nulls,
    array_not_empty,
)
from visions.backends.numpy.test_utils import coercion_map, coercion_map_test
from visions.backends.python.types.boolean import get_boolean_coercions
from visions.backends.shared.nan_handling import nan_mask
from visions.types.boolean import Boolean
from visions.types.object import Object
from visions.types.string import String

string_coercions = get_boolean_coercions("en")


@Boolean.register_relationship(Object, np.ndarray)
@array_handle_nulls
def object_is_boolean(array: np.ndarray, state: dict) -> bool:
    return all_type(array, bool)


@Boolean.register_transformer(Object, np.ndarray)
def object_to_boolean(array: np.ndarray, state: dict) -> np.ndarray:
    return array


@Boolean.register_relationship(String, np.ndarray)
def string_is_boolean(array: np.ndarray, state: dict) -> bool:
    try:
        mask = nan_mask(array)
        # TODO: Nan handling not implemented for generators yet
        val_generator = np.array([val.lower() for val in array[mask]])
        return coercion_map_test(string_coercions)(val_generator, state)
    except (ValueError, TypeError, AttributeError):
        return False


@Boolean.register_transformer(String, np.ndarray)
def string_to_boolean(array: np.ndarray, state: dict) -> np.ndarray:
    array = array.copy()
    mask = nan_mask(array)
    # TODO: Nan handling not implemented for generators yet
    val_generator = np.array([val.lower() for val in array[mask]])
    array[mask] = object_to_boolean(
        coercion_map(string_coercions)(val_generator), state
    )
    return array


@Boolean.contains_op.register
@array_handle_nulls
@array_not_empty
def boolean_contains(array: np.ndarray, state: dict) -> bool:
    if np.issubdtype(array.dtype, np.bool_):
        return True

    return all_type(array, bool)



from typing import Dict, Sequence
from urllib.parse import urlparse

import numpy as np


def get_sequences() -> Dict[str, Sequence]:
    sequences = {
        "complex_series_float": [
            complex(0, 0),
            complex(1, 0),
            complex(3, 0),
            complex(-1, 0),
        ],
        "url_nan_series": [
            urlparse("http://www.cwi.nl:80/%7Eguido/Python.html"),
            urlparse("https://github.com/dylan-profiling/hurricane"),
            np.nan,
        ],
        "mixed": [True, False, np.nan],
        "float_nan_series": [1.0, 2.5, np.nan],
        "float_series5": [np.nan, 1.2],
        "float_with_inf": [np.inf, np.NINF, np.PINF, 1000000.0, 5.5],
        "inf_series": [np.inf, np.NINF, np.Infinity, np.PINF],
        "int_nan_series": [1, 2, np.nan],
        "nan_series": [np.nan],
        "nan_series_2": [np.nan, np.nan, np.nan, np.nan],
        "string_num_nan": ["1.0", "2.0", np.nan],
        "string_with_sep_num_nan": ["1,000.0", "2.1", np.nan],
        "string_flt_nan": ["1.0", "45.67", np.nan],
        "string_str_nan": [
            "I was only robbing the register,",
            "I hope you understand",
            "One of us had better call up the cops",
            "In the hot New Jersey night",
            np.nan,
        ],
        "float_series3": np.array([1.2, 2, 3, 4], dtype=np.float64),
        "np_uint32": np.array([1, 2, 3, 4], dtype=np.uint32),
        "string_np_unicode_series": np.array(["upper", "hall"], dtype=np.unicode_),
        "complex_series": [
            complex(0, 0),
            complex(1, 2),
            complex(3, -1),
        ],
        "bool_series3": np.array([1, 0, 0, 1], dtype=np.bool_),
        "complex_series_nan": [complex(0, 0), complex(1, 2), complex(3, -1), None],
        "complex_series_nan_2": [
            complex(0, 0),
            complex(1, 2),
            complex(3, -1),
            np.nan,
        ],
        "complex_series_py_nan": [
            complex(0, 0),
            complex(1, 2),
            complex(3, -1),
            np.nan,
        ],
    }
    return sequences



import numpy as np
import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas import test_utils
from visions.backends.pandas.series_utils import (
    series_handle_nulls,
    series_not_empty,
    series_not_sparse,
)
from visions.types.complex import Complex
from visions.types.float import Float
from visions.types.string import String
from visions.utils.warning_handling import suppress_warnings


def test_string_leading_zeros(series: pd.Series, coerced_series: pd.Series):
    if coerced_series.hasnans:
        notna = coerced_series.notna()
        coerced_series = coerced_series[notna]

        if coerced_series.empty:
            return False
        series = series[notna]
    return not any(s[0] == "0" for s in series[coerced_series > 1])


@Float.register_relationship(String, pd.Series)
@series_handle_nulls
def string_is_float(series: pd.Series, state: dict) -> bool:
    coerced_series = test_utils.option_coercion_evaluator(lambda s: s.astype(float))(
        series
    )

    return (
        coerced_series is not None
        and float_contains(coerced_series, state)
        and test_string_leading_zeros(series, coerced_series)
    )


@Float.register_transformer(String, pd.Series)
def string_to_float(series: pd.Series, state: dict) -> pd.Series:
    # Slightly faster to check for the character if it's not present than to
    # attempt the replacement
    # if any("," in x for x in series):
    #     series = series.str.replace(",", "")
    return series.astype(float)


@Float.register_relationship(Complex, pd.Series)
def complex_is_float(series: pd.Series, state: dict) -> bool:
    return all(np.imag(series.values) == 0)


@Float.register_transformer(Complex, pd.Series)
def complex_to_float(series: pd.Series, state: dict) -> pd.Series:
    return suppress_warnings(lambda s: s.astype(float))(series)


@Float.contains_op.register
@series_not_sparse
@series_handle_nulls
@series_not_empty
def float_contains(series: pd.Series, state: dict) -> bool:
    return pdt.is_float_dtype(series)



from typing import Sequence

from visions.types.count import Count


@Count.contains_op.register
def count_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(value, int) and value >= 0 for value in sequence)



# from visions import String
# from visions.relations.relations import InferenceRelation
#
#
# def string_to_categorical_distinct_count() -> InferenceRelation:
#     """Convert string to categorical when it has fewer than 50% unique values.
#
#     Returns:
#         relation
#     """
#     # TODO: only when not any other string relation (either exclude others or have ordering and evaluate last)
#     return InferenceRelation(
#         relationship=lambda s, state: s.nunique() / len(s) < 0.5,
#         transformer=lambda s: s.astype("category"),
#         related_type=String,
#     )



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, TypeRelation
from visions.types.object import Object
from visions.types.type import VisionsBaseType


class Time(VisionsBaseType):
    """**Time** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> import datetime
        >>> import visions
        >>> x = [datetime.time(10, 8, 4), datetime.time(21, 17, 0)]
        >>> x in visions.Time
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [IdentityRelation(Object)]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



import functools
from typing import Callable

import pandas as pd


# For future reference: get the dtype from the subtype when the series is sparse
def series_handle_sparse_dtype(fn: Callable[..., bool]) -> Callable[..., bool]:
    """Decorator to include the dtype of a sparse subtype."""

    @functools.wraps(fn)
    def inner(series: pd.Series, state: dict, *args, **kwargs) -> bool:
        if isinstance(series.dtype, pd.SparseDtype):
            dtype = series.dtype.subtype
        else:
            dtype = series.dtype
        state["dtype"] = dtype

        return fn(series, state, *args, **kwargs)

    return inner


def series_handle_nulls(fn: Callable[..., bool]) -> Callable[..., bool]:
    """Decorator for nullable series"""

    @functools.wraps(fn)
    def inner(series: pd.Series, *args, **kwargs) -> bool:
        if series.hasnans:
            series = series.dropna()
            # TODO: use series_not_empty?
            if series.empty:
                return False

        return fn(series, *args, **kwargs)

    return inner


def series_not_sparse(fn: Callable[..., bool]) -> Callable[..., bool]:
    """Decorator to exclude sparse series"""

    @functools.wraps(fn)
    def inner(series: pd.Series, *args, **kwargs) -> bool:
        if isinstance(series, pd.SparseDtype):
            return False
        return fn(series, *args, **kwargs)

    return inner


def series_not_empty(fn: Callable[..., bool]) -> Callable[..., bool]:
    """Decorator to exclude empty series"""

    @functools.wraps(fn)
    def inner(series: pd.Series, *args, **kwargs) -> bool:
        if series.empty:
            return False
        return fn(series, *args, **kwargs)

    return inner


# TODO: What is the type signature on is_method????
def _contains_instance_attrs(
    series: pd.Series, is_method, class_name: str, attrs: list, sample_size: int = 1
) -> bool:
    # TODO: user configurable .head or .sample
    # TODO: performance testing for series[0], series.iloc[0], series.head, series.sample
    if not all(is_method(x, class_name) for x in series.head(sample_size)):
        return False

    try:
        return all(all(hasattr(x, attr) for attr in attrs) for x in series)
    except AttributeError:
        return False


# TODO: What is the type signature on class_name????
def class_name_attrs(
    series: pd.Series, class_name, attrs: list, sample_size: int = 1
) -> bool:
    def func(instance, class_name):
        return instance.__class__.__name__ == class_name.__name__

    return _contains_instance_attrs(series, func, class_name, attrs, sample_size)


# TODO: What is the type signature on class_name????
def isinstance_attrs(
    series: pd.Series, class_name, attrs: list, sample_size: int = 1
) -> bool:
    return _contains_instance_attrs(series, isinstance, class_name, attrs, sample_size)



"""This module contains (predefined) relations."""

from visions.relations.relations import (
    IdentityRelation,
    InferenceRelation,
    TypeRelation,
)

__all__ = [
    "IdentityRelation",
    "InferenceRelation",
    "TypeRelation",
]



from visions.typesets.complete_set import CompleteSet
from visions.typesets.geometry_set import GeometrySet
from visions.typesets.standard_set import StandardSet
from visions.typesets.typeset import VisionsTypeset

__all__ = ["VisionsTypeset", "CompleteSet", "StandardSet", "GeometrySet"]



from typing import Sequence

from visions.backends.python.series_utils import (
    sequence_handle_none,
    sequence_not_empty,
)
from visions.types.object import Object


@Object.contains_op.register
@sequence_not_empty
@sequence_handle_none
def object_contains(sequence: Sequence, state: dict) -> bool:
    return any(not isinstance(value, (float, bool, int, complex)) for value in sequence)



from typing import Sequence

from visions.backends.python.series_utils import sequence_not_empty
from visions.types.float import Float
from visions.types.integer import Integer


@Integer.register_relationship(Float, Sequence)
def float_is_int(sequence: Sequence, state: dict) -> bool:
    try:
        return all(int(value) == value for value in sequence)
    except (ValueError, TypeError, OverflowError):
        return False


@Integer.register_transformer(Float, Sequence)
def float_to_int(sequence: Sequence, state: dict) -> Sequence:
    return tuple(map(int, sequence))


@Integer.contains_op.register
@sequence_not_empty
def integer_contains(sequence: Sequence, state: dict) -> bool:
    return all(
        isinstance(value, int) and not isinstance(value, bool) for value in sequence
    )



import numpy as np
import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas.series_utils import (
    series_handle_nulls,
    series_not_empty,
    series_not_sparse,
)
from visions.types.float import Float
from visions.types.integer import Integer


@Integer.register_relationship(Float, pd.Series)
@series_handle_nulls
def float_is_integer(series: pd.Series, state: dict) -> bool:
    def check_equality(series):
        try:
            if not np.isfinite(series).all():
                return False
            return series.eq(series.astype(int)).all()
        except (ValueError, TypeError, AttributeError):
            return False

    return check_equality(series)


@Integer.register_transformer(Float, pd.Series)
def float_to_integer(series: pd.Series, state: dict) -> pd.Series:
    dtype = "Int64" if series.hasnans else np.int64
    return series.astype(dtype)


@Integer.contains_op.register
@series_not_sparse
@series_not_empty
def integer_contains(series: pd.Series, state: dict) -> bool:
    return pdt.is_integer_dtype(series)



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, TypeRelation
from visions.types.file import File
from visions.types.type import VisionsBaseType


class Image(VisionsBaseType):
    """**Image** implementation of :class:`visions.types.type.VisionsBaseType`.
    (i.e. series with all image files)

    Examples:
        >>> from pathlib import Path
        >>> import visions
        >>> x = [Path('/home/user/file.png'), Path('/home/user/test2.jpg')]
        >>> x in visions.Image
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [IdentityRelation(File)]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



from datetime import datetime
from functools import partial

import numpy as np
import pandas as pd

from visions.backends.numpy import test_utils
from visions.backends.numpy.array_utils import array_handle_nulls, array_not_empty
from visions.backends.pandas.types.date_time import pandas_infer_datetime
from visions.types import DateTime, String


@DateTime.register_relationship(String, np.ndarray)
@array_handle_nulls
def string_is_datetime(array: np.ndarray, state: dict) -> bool:
    exceptions = [OverflowError, TypeError]

    if len(array) == 0:
        return False

    coerced_array = test_utils.option_coercion_evaluator(
        partial(string_to_datetime, state=state), exceptions
    )(array)

    if coerced_array is None:
        return False
    elif np.isnat(coerced_array).any():
        return False

    return True


@DateTime.register_transformer(String, np.ndarray)
def string_to_datetime(array: np.ndarray, state: dict) -> np.ndarray:
    # return array.astype(np.datetime64)
    return pandas_infer_datetime(pd.Series(array), state).to_numpy()


@DateTime.contains_op.register
@array_handle_nulls
@array_not_empty
def datetime_contains(array: np.ndarray, state: dict) -> bool:
    if np.issubdtype(array.dtype, np.datetime64):
        return True

    return all(isinstance(v, datetime) for v in array)
